from .config import CONFIG
from .formatters import FORMATTING_PASSES
from .helpers import color_diff, print_error
from .helpers import debug1, debug2, debug3, debug4
from pathlib import Path
from redbaron import RedBaron
import difflib


def process_file(fname: Path, *, interactive: bool = False, accept: bool = False, dry: bool = False, no_diff: bool = False):
    assert sum(map(bool, [interactive, accept, dry])) == 1
    if no_diff: assert dry

    with fname.open() as f:
        orig_src = f.read()
        try:
            red = RedBaron(orig_src)
        #except baron.parser.ParsingError as e:
        except Exception as e:
            print_error(f"== {e.__class__.__name__}: ==\n{e}\n")
            return
        assert red.dumps() == orig_src

    def ask_and_write(red: RedBaron) -> bool:
        new_src = red.dumps()
        if dry:
            if no_diff:
                print(new_src)
            else:
                view_diff(fname, new_src)
            return True
        elif accept or view_diff_ask_accept(fname, new_src):
            with open(fname, "w") as f:
                f.write(new_src)
            return True
        return False

    for formatting_pass in FORMATTING_PASSES:
        #name = formatting_pass.__name__.capitalize().replace("__", " - ").replace("_", " ")
        #print("    -", name)

        #print("    -", name, end="\r")
        #print()
        #sys.stdout.write("\033[K") # clear to end of line

        #debug3(red)
        red = formatting_pass(red, interactive=ask_and_write if interactive else None)
        ask_and_write(red)
        #debug3(red)

def view_diff(fname: Path, new_src: str):
    #if shutil.which("diff"):
    #    subprocess.run([
    #            "diff",
    #            "-Bwu",
    #            "--color",
    #            fname,
    #            "-",
    #        ], input=new_src, text=True)
    #    #TODO: check if diff returned success due to no difference
    #elif shutil.which("git"):
    #    subprocess.run([
    #            "git",
    #            "diff",
    #            "-U0",
    #            #"--word-diff",
    #            #"--word-diff-regex=[<]|[>]|[^[:space:]]",
    #            "--no-index",
    #            "--", fname, "-"
    #        ], input=new_src, text=True)
    #    #TODO: check if diff returned success due to no difference
    #else:

    with fname.open() as f:
        original_file = f.read()

    diff = list(difflib.unified_diff(
        original_file.splitlines(keepends=True),
        new_src      .splitlines(keepends=True),
        fromfile = str(fname),
        tofile   = str(fname),
    ))
    if diff:
        print(
            *color_diff(diff, force=CONFIG.color),
            sep="",
        )
        return True
    else:
        return False

def view_diff_ask_accept(fname: Path, new_src: str) -> bool:
    if view_diff(fname, new_src):
        return input("Apply changes? [y/N] ").lower().startswith("y")
    else:
        return False



from .config import CONFIG
from .helpers import is_self_arg, make_comma, make_whitespace, make_multiline_comments
from .helpers import elementwise_max
from .helpers import debug1, debug2, debug3, debug4
from redbaron import RedBaron
from typing import Callable, Optional
import json


FormattingPass      = Callable[[RedBaron, ...], RedBaron]
InteractiveCallable = Optional[Callable[[RedBaron], bool]]

FORMATTING_PASSES: list[FormattingPass] = []
def _formatting_pass(func):
    FORMATTING_PASSES.append(func)
    return func

# === formatting passes

@_formatting_pass
def tabulate_function_definitions(red: RedBaron, *, interactive: InteractiveCallable) -> RedBaron:
    for i, node in enumerate(red.find_all("def")):
        backup_node = node.copy()

        #debug1(node)

        # read/derive indent levels
        indent_def = str(node.get_indentation_node() or "\n")
        indent_body = str(node.value[0].get_indentation_node() or "\n")
        indent_diff = indent_body.removeprefix(indent_def)
        if indent_diff:
            #debug1(indent_def)
            #debug1(indent_body)
            assert node.value.node_list[0].type == "endl"
        else:
            assert node.value.node_list[0].type != "endl"
            indent_diff = " " * CONFIG.tab_width # TODO: support tabs
            indent_body += indent_diff
            if node.value.node_list[0].type == "space":
                node.value.node_list[0].replace(make_whitespace(indent_body))
            else:
                node.value.node_list.insert(0, make_whitespace(indent_body))
        indent_args = indent_body + indent_diff

        # store some user comments before we wipe them below
        def_comment_nodes_head = [
            *(i.copy() for i in node.async_formatting .find_all("comment")),
            *(i.copy() for i in node.first_formatting .find_all("comment")),
            *(i.copy() for i in node.second_formatting.find_all("comment")),
            *(i.copy() for i in node.third_formatting .find_all("comment")),
        ]
        def_comment_nodes_tail = [
            *(i.copy() for i in node.fourth_formatting.find_all("comment")),
            *(i.copy() for i in node.return_annotation_first_formatting .find_all("comment")),
            *(i.copy() for i in node.return_annotation_second_formatting.find_all("comment")),
            *(i.copy() for i in node.fifth_formatting .find_all("comment")),
            *(i.copy() for i in node.sixth_formatting .find_all("comment")),
        ]
        try:
            has_def_argument_comments = bool(node.arguments[is_self_arg(node.arguments[0]):].find("comment"))
        except IndexError:
            has_def_argument_comments = False

        # set fixed whitespace
        node.async_formatting  = " "
        node.first_formatting  = " "                   # between def and name
        node.second_formatting = ""                    # between name and (
        node.third_formatting  = ""                    # between ( and first token
        node.fourth_formatting = ""                    # between last token and )
        node.return_annotation_first_formatting  = " " # before ->
        node.return_annotation_second_formatting = " " # after ->
        node.fifth_formatting  = ""                    # before :
        node.sixth_formatting  = ""                    # after :

        #debug2(node)
        #debug1(node.arguments)
        #debug2(node.arguments)

        supported_arg_node_types = {
            "def_argument",        # foo   : bar = baz
            "list_argument",       # *foo  : bar
            "dict_argument",       # **foo : bar
            "kwargs_only_marker",  # *
            #"args_only_marker,    # /, not implemented in baron
        }
        assert all(arg_node.type in supported_arg_node_types for arg_node in node.arguments), debug4(node.arguments)
        n_items = bool(node.return_annotation) + sum(
            1 + bool(arg_node.annotation) + bool(arg_node.value)
            if arg_node.type != "kwargs_only_marker" else
            1
            for arg_node in node.arguments
            if not is_self_arg(arg_node)
        )
        if n_items >= 5 or "\n" in node.arguments.dumps():
            node.third_formatting = indent_args

            l1m, l2m, l3m = elementwise_max(
                (
                    len(str(arg_node.target     or "")),
                    len(str(arg_node.annotation or "")),
                    len(str(arg_node.value      or "")),
                )
                if arg_node.type == "def_argument" else
                (
                    len(str(arg_node.value      or "")) + 1 + (arg_node.type == "dict_argument"),
                    len(str(arg_node.annotation or "")),
                    0,
                )
                for arg_node in node.arguments
                if not is_self_arg(arg_node)
                    and not arg_node.type == "kwargs_only_marker"
            )

            #if node.arguments.node_list.find("comment") and node..find("comment"):

            # ensure trailing comma
            if not node.arguments.has_trailing:
                node.arguments.node_list.append(make_comma())

            for arg_node in node.arguments:
                assert arg_node.type in supported_arg_node_types, arg_node.type

                if arg_node.type in {
                        "list_argument",
                        "dict_argument",
                        "kwargs_only_marker",
                        }:
                    arg_node.formatting = ""

                comma_node = arg_node.next
                assert comma_node
                assert comma_node.type == "comma", comma_node.type

                if is_self_arg(arg_node): # special case for "self" and "cls"
                    node.third_formatting                 = ""
                    arg_node.first_formatting             = ""
                    arg_node.second_formatting            = ""
                    arg_node.annotation_first_formatting  = ""
                    arg_node.annotation_second_formatting = ""

                    comment_nodes = [*def_comment_nodes_head, *comma_node.find_all("comment")]
                elif arg_node.type != "kwargs_only_marker":
                    l1, l2, l3 = (
                        (
                            len(str(arg_node.target     or "")),
                            len(str(arg_node.annotation or "")),
                            len(str(arg_node.value      or "")),
                        )
                        if arg_node.type == "def_argument" else
                        (
                            len(str(arg_node.value      or "")) + 1 + (arg_node.type == "dict_argument"),
                            len(str(arg_node.annotation or "")),
                            0,
                        )
                    )

                    # annotations
                    arg_node.annotation_first_formatting  = " " * (1 + l1m - l1)
                    arg_node.annotation_second_formatting = " "

                    # default values
                    if arg_node.type == "def_argument":
                        arg_node.first_formatting         = " " * (1 + l2m - l2 + (l1m - l1 + 3)*(not l2))
                        arg_node.second_formatting        = " "

                    comment_nodes = comma_node.find_all("comment")

                # comments and endl
                comma_node.first_formatting  = ""
                if comment_nodes:
                    sep  = indent_args + " " * (l1m + 3 + l2m + 3 + l3m + 1 + 2)
                    comma_column = comma_node.absolute_bounding_box.top_left.column

                    do_dedent_top_comment = is_self_arg(arg_node) \
                        and (len(comment_nodes) > 1 or not has_def_argument_comments)

                    if do_dedent_top_comment:
                        sep = indent_args + " " * (comma_column + 2 - len(indent_args.lstrip("\r\n")))

                    lead = " " * (len(sep.lstrip("\r\n")) - comma_column)
                    #lead = " " * (2 + l3m - l3)

                    # eww
                    if do_dedent_top_comment:
                        comma_column -= indent_diff.count("\t") * (CONFIG.tab_width-1) * 2
                        sep = indent_args + " " * (comma_column + 2 - len(indent_args.lstrip("\r\n")))

                    comma_node.second_formatting.clear()
                    comma_node.second_formatting.extend(
                        make_multiline_comments(comment_nodes, lead, sep, indent_args)
                    )
                else:
                    comma_node.second_formatting = indent_args

            # final comments
            if def_comment_nodes_tail:
                assert node.value.node_list[0].type == "endl", json.dumps(node.value.node_list[0].fst(), indent=4)
                column = node.value.node_list[0].absolute_bounding_box.top_left.column
                sep    = indent_args + " " * (l1m + 3 + l2m + 3 + l3m + 1 + 2)
                lead   = " " * (len(sep.lstrip("\r\n")) - column)

                if not has_def_argument_comments:
                    sep = sep[:-(len(lead) - 2)]
                    lead = " " * 2

                #node.sixth_formatting.clear()
                node.sixth_formatting.extend(
                    make_multiline_comments(def_comment_nodes_tail, lead, sep, indent_args)
                )
        else:
            node.replace(backup_node)
            continue

        if interactive:
            if node.dumps() != backup_node.dumps():
                if not interactive(red):
                    # rollback
                    node.replace(backup_node)

    return red


# TODO: def tabulate_function_calls

# TODO: def tabulate_dictionaries

# TODO: def tabulate_assignment_blocks



from .config import CONFIG
from functools import reduce
from colorama import Fore
from redbaron import nodes
from typing import Union, Iterable, Any
import json
import os
import sys

# baron helpers:

def is_self_arg(def_argument_node: nodes.DefArgumentNode) -> bool:
    return (def_argument_node.index_on_parent == 0
        and def_argument_node.type            == "def_argument"
        and def_argument_node.target.type     == "name"
        and def_argument_node.target.value    in {"self", "cls"}
        and not def_argument_node.annotation
    )

def make_comma() -> nodes.CommaNode:
    return nodes.CommaNode({
        "type"              : "comma",
        "first_formatting"  :  "",
        "second_formatting" :  "",
    })

def make_whitespace(value) -> Union[nodes.SpaceNode, nodes.EndlNode]:
    endl, indent = value.rstrip(" \t"), value.lstrip("\r\n")
    if endl:
        return nodes.EndlNode({
            "type"       : "endl",
            "value"      :  endl,
            "indent"     :  indent,
            "formatting" : [],
        })
    else:
        return nodes.SpaceNode({
            "type"  : "space",
            "value" :  indent,
        })

def make_multiline_comments(comment_nodes: list[nodes.CommentNode], whitespace_head: str, whitespace_sep: str, whitespace_tail: str) -> Iterable[nodes.Node]:
    yield make_whitespace(whitespace_head)
    for i, comment_node in enumerate(comment_nodes):
        if i != 0:
            yield make_whitespace(whitespace_sep)
        yield comment_node
    yield make_whitespace(whitespace_tail)

# pretty-print helpers

def color_diff(diff: list[str], force: bool = False) -> Iterable[str]:
    if not force and not os.isatty(1):
        yield from diff
    else:
        for line in diff:
            if line.startswith('+++'):
                yield f"{ Fore.CYAN   }{ line }{ Fore.RESET }"
            elif line.startswith('---'):
                yield f"{ Fore.BLUE   }{ line }{ Fore.RESET }"
            elif line.startswith('+'):
                yield f"{ Fore.GREEN }{ line }{ Fore.RESET }"
            elif line.startswith('-'):
                yield f"{ Fore.RED   }{ line }{ Fore.RESET }"
            elif line.startswith('^'):
                yield f"{ Fore.BLUE  }{ line }{ Fore.RESET }"
            elif line.startswith('@@') and line.rstrip().endswith("@@"):
                yield f"{ Fore.MAGENTA  }{ line }{ Fore.RESET }"
            else:
                yield line

def print_error(* a, ** kw):
    if os.isatty(2) or CONFIG.color:
        print(Fore.RED, end="", file=sys.stderr)
        print(*a, **kw, file=sys.stderr)
        print(Fore.RESET, end="", file=sys.stderr)
        sys.stderr.flush()
    else:
        print(*a, **kw, file=sys.stderr)


# itertools

def elementwise_max(iterable: Iterable):
    return reduce(lambda xs, ys: [*map(max, zip(xs, ys))], iterable)


# print debugging

def debug1(*a, **kw):
    return print(*map(repr, a), **kw)

def debug2(obj: Any):
    return debug1(dir(obj))

def debug3(obj: nodes.Node):
    return print(debug4(obj))

def debug4(obj: nodes.Node):
    return json.dumps(obj.fst(), indent=4)



#!/usr/bin/env python3
from .config import CONFIG
from .runner import process_file
from argparse import ArgumentParser
from itertools import chain
from pathlib import Path
import argparse


__doc__ = """
Possible invocations:

    fix-my-functions
    python -m fix_my_functions
    poetry run fix-my-functions
    poetry run python -m fix_my_functions
"""


def make_argparser(parser = None) -> ArgumentParser:
    if parser is None:
        parser = ArgumentParser()

    parser_input = parser.add_argument_group("input")
    parser_input.add_argument("files", type=str, nargs="+",
        help="Files to format.")
    parser_input.add_argument("-r", "--recursive", action="store_true",
        help="Recursively format .py files in folders")
    parser_input_mutex = parser_input.add_mutually_exclusive_group(required=True)
    parser_input_mutex.add_argument("-a", "--accept", action="store_true",
        help="Accept and apply all changes.")
    parser_input_mutex.add_argument("-i", "--interactive", action="store_true",
        help="Ask to apply each change.")
    parser_input_mutex.add_argument("-d", "--dry", action="store_true",
        help="Only show unified diff.")

    parser_input.add_argument("--no-diff", action="store_true",
        help="Only show raw output instead of unififed diff, requires --dry.")

    parser = CONFIG.make_argparser(parser)

    class CustomFormatter(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        ): pass
    parser.formatter_class = CustomFormatter
    parser.description = __doc__

    return parser

def main(parser = make_argparser()):
    args = parser.parse_args() # may terminate on --help and such
    if args.no_diff and not args.dry:
        print("ERROR: --no-diff requires --dry")
        return 1

    CONFIG.parse_args(args)

    files = [*map(Path, args.files)]

    if args.recursive:
        files = sorted(set(chain(*(
            file.rglob("*.py")
            if file.is_dir() else
            [file]
            for file in files
        ))))

    for file in files:
        if file.is_dir():
            print(f"Skipping {file}, because it is a directory")
            continue

        if len(files) > 1:
            print(file)

        process_file(file,
            dry         = args.dry,
            interactive = args.interactive,
            accept      = args.accept,
            no_diff     = args.no_diff,
        )

if __name__ == "__main__":
    exit(main() or 0)



from argparse import ArgumentParser, Namespace


class Config:
    tab_width : int  = 4
    color     : bool = False

    def make_argparser(self, parser: ArgumentParser = None) -> ArgumentParser:
        if parser is None:
            parser = ArgumentParser()

        parser_config = parser.add_argument_group("config")

        parser_config.add_argument("--tab-width", type=int, default=self.tab_width,
            help="Which tab width to use when calculating the column position for \\t.")

        parser_config.add_argument("--color-output", action="store_true",
            help="Force colored output.")

        return parser

    def parse_args(self, args: Namespace):
        self.tab_width = args.tab_width
        self.color     = args.color_output


# singleton
CONFIG = Config()
del Config

# This enables all who import CONFIG to make changes, visible to all as they
# share the same reference
